console.log('in fs_facade');

var fs = require('fs');
var path = require('path');
var os = require('os');

var dict = global.require('dict');
var q = global.require('q');
var handlebars = global.require('handlebars');

var module_manager;
try {
    module_manager = require('ljswitchboard-module_manager');
} catch(err) {
    // console.warn('Error!! requiring module_manager', err);
}
try {
    module_manager = global.require('ljswitchboard-module_manager');
} catch(err) {
    console.warn('Error!! global.requiring module_manager', err);
}

var FS_FACADE_TEMPLATE_CACHE = dict();
exports.numCachedTemplates = function() {
    return FS_FACADE_TEMPLATE_CACHE.size;
};

exports.clearCache = function() {
    FS_FACADE_TEMPLATE_CACHE = dict();
}

/**
 * Render a Handlebars template.
 *
 * @param {String} location The location of the template to render.
 * @param {Object} context The context to render the template with.
 * @param {function} onError The function to call if an error is encountered
 *      while rendering the template. Should take a single argument which
 *      would be the error.
 * @param {function} onSuccess The function to call after the template is
 *      successfully rendered. Should take a single argument which would
 *      be the String rendred html.
**/
exports.renderTemplate = function(location, context, onError, onSuccess) {
    if (FS_FACADE_TEMPLATE_CACHE.has(location)) {
        var curTemplate = FS_FACADE_TEMPLATE_CACHE.get(location);
        onSuccess(curTemplate(context));
    } else {
        fs.exists(location, function(exists) {
            if (exists) {
                fs.readFile(location, 'utf8',
                    function (error, template) {
                        if (error) {
                            onError(error);
                        } else {
                            var curTemplate = handlebars.compile(template);
                            FS_FACADE_TEMPLATE_CACHE.set(location,curTemplate);
                            onSuccess(curTemplate(context));
                        }
                    }
                );
            } else {
                onError(new Error('Template ' + location + ' could not be found.'));
            }
        });
    }
};

var renderTemplateData = function(template, context) {
	var defered = q.defer();
	var data = "";
	try {
		var curTemplate = handlebars.compile(template);
		data = curTemplate(context);
		defered.resolve(data);
	} catch(err) {
		defered.resolve('');
	}
	return defered.promise;
};
exports.renderTemplateData = renderTemplateData;

exports.renderCachedTemplateData = function(key, template, context) {
	var defered = q.defer();
	var curTemplate;
	if(FS_FACADE_TEMPLATE_CACHE.has(key)) {
		curTemplate = FS_FACADE_TEMPLATE_CACHE.get(key);
		defered.resolve(curTemplate(context));
	} else {
		curTemplate = handlebars.compile(template);
		FS_FACADE_TEMPLATE_CACHE.set(key, curTemplate);
		defered.resolve(curTemplate(context));
	}
	return defered.promise;
};

/*
Required functions for lua debugger module to work:

Function calls in controller.js:
1. fs_facade.getFileLoadID()
    Returns a constant "#file-dialog-hidden" (handles the file IO window)
2. fs_facade.readModuleFile(fileLocation, onError, onSuccess)
    Appends a file path with the location of all module files, checks to see if
    it is there and then reads it.
Function calls in helperFunctions.js:
1. fs_facade.readModuleFile(fileLocation, onError, onSuccess)
    Appends a file path with the location of all module files, checks to see if
    it is there and then reads it.

Function calls in luaDeviceController.js:
1. fs_facade.saveDataToFile(filePath, scriptData, onError, onSuccess);
    Checks for a file and writes data to it.  (This may have bugs according to 
    caleb's testing.)
2. fs_facade.getFileSaveAsID()
    Returns a constant "#file-save-dialog-hidden"
3. fs_facade.getDefaultFilePath()
    Returns a constant that is platform based.
4. fs_facade.loadFile(filePath, onError, onSuccess)
    Is a wrapper around the fs.exists & fs.readFile functions.
5. fs_facade.readModuleFile(fileLocation, onError, onSuccess)
    Appends a file path with the location of all module files, checks to see if
    it is there and then reads it.

*/

var fileSaveAsID = "#file-save-dialog-hidden";
var getFileLoadID = "#file-dialog-hidden";


var windowsDefaultFilePath = 'C:\\Windows';

if(process.env.HOME) {
    windowsDefaultFilePath = process.env.HOME;
}

var defaultFilePath = {
    'linux': '/home/path',
    'linux2': '/home/path',
    'sunos': '/home/path',
    'solaris': '/home/path',
    'freebsd': '/home/path',
    'openbsd': '/home/path',
    'darwin': process.env.HOME,
    'mac': process.env.HOME,
    'win32': windowsDefaultFilePath,
    'win64': windowsDefaultFilePath
}[process.platform];
exports.testFunction = function() {
    return "testing...";
};
exports.getFileSaveAsID = function() {
    return fileSaveAsID;
};
exports.getFileLoadID = function() {
    return getFileLoadID;
};
exports.getDefaultFilePath = function() {
    return defaultFilePath;
};

exports.saveDataToFile = function(location, data, onError, onSuccess) {
    fs.exists(location, function(exists) {
        fs.writeFile(
            location,
            data,
            function (err) {
                if (err) {
                    onError(err);
                } else {
                    onSuccess();
                }
            }
        );
    });
};

exports.loadFile = function(location, onError, onSuccess) {
    fs.exists(location, function(exists) {
        if (exists) {
            fs.readFile(location, 'utf8',
                function (error, contents) {
                    if (error) {
                        fs.readFile(location, 'utf8',
                            function (error, contents) {
                                if (error) {
                                    onError(error);
                                } else {
                                    onSuccess(contents);
                                }
                            }
                        );
                    } else {
                        onSuccess(contents);
                    }
                }
            );
        } else {
            var error = new Error('Could not find File at ' + location + '.');
            onError(error);
        }
    });
};

exports.getExternalURI = function(filePath) {
    var modulesDir = module_manager.getModulesDirectory();
    return path.normalize(path.join(modulesDir, filePath));
};
exports.readModuleFile = function(filePath, onError, onSuccess) {
    var fullPath = exports.getExternalURI(filePath);
    fs.exists(fullPath, function(exists) {
        if (exists) {
            fs.readFile(fullPath, 'utf8',
                function (error, contents) {
                    if (error) {
                        onError(error);
                    } else {
                        onSuccess(contents);
                    }
                }
            );
        } else {
            var error = new Error(
                'Could not find modules info at ' + fullPath + '.'
            );
            onError(error);
        }
    });
};